@@ -73,7 +73,21 @@ typedef struct _UpdateTrack{
     GtkTreeIter it;
 }UpdateTrack;
 
-static void send_notifcation( const gchar *artist, const gchar* title );
+typedef struct _TrackProperties{
+    const char *artist;
+    const char *album;
+    const char *title;
+    const char *url;
+    const char *mime;
+    const char *comment;
+    int32_t duration;
+    int32_t isvbr;
+    int32_t bitrate;
+    int32_t size;
+}TrackProperties;
+
+static void 	send_notifcation			( const gchar *artist, const gchar* title );
+
 #ifdef HAVE_LIBNOTIFY
 static LXMusic_Notification *lxmusic_notification = NULL;
 #endif 
@@ -100,14 +114,15 @@ static GtkWidget *rm_from_pl_menu = NULL;
 static char* cur_playlist = NULL;
 static GSList* all_playlists = NULL;
 
-static GQueue* pending_update_tracks = NULL;
+/* to be updated by update_track */
+static GHashTable *update_tracks;
+
 static GtkListStore* list_store = NULL;
 
 static int32_t playback_status = 0;
 static uint32_t play_time = 0;
 static int32_t cur_track_duration = 0;
 static int32_t cur_track_id = 0;
-static GtkTreeIter cur_track_iter = {0};
 
 static int repeat_mode = REPEAT_NONE;
 
@@ -130,7 +145,11 @@ static int win_height = 320;
 static int win_xpos = 0;
 static int win_ypos = 0;
 
-void on_play_btn_clicked(GtkButton* btn, gpointer user_data);
+void 			on_locate_cur_track	(GtkAction* act, gpointer user_data);
+void 			on_play_btn_clicked	(GtkButton* btn, gpointer user_data);
+
+static GtkTreeIter	get_current_track_iter	();
+static gboolean		get_track_properties 	(xmmsv_t *value, TrackProperties *properties);
 
 static void load_config()
 {
@@ -185,19 +204,9 @@ static void save_config()
     }
 }
 
-static void free_update_track( UpdateTrack* ut )
-{
-    g_slice_free(UpdateTrack, ut);
-}
-
 static void cancel_pending_update_tracks()
 {
-    /* g_debug("try to cancel"); */
-    if( ! g_queue_is_empty(pending_update_tracks) )
-    {
-        g_queue_foreach(pending_update_tracks, (GFunc)free_update_track, NULL);
-        g_queue_clear(pending_update_tracks);
-    }
+    g_hash_table_remove_all( update_tracks );
 }
 
 static int on_xmms_quit(xmmsv_t *value, void *user_data)
@@ -305,7 +314,7 @@ void on_pref_output_plugin_changed(GtkTable* table, GtkComboBox* output)
     n_configs = g_list_length( plugin->config );
 
     /* destroy previous configuration widgets */
-    gtk_container_foreach( GTK_CONTAINER( table ), gtk_widget_destroy, NULL );
+    gtk_container_foreach( GTK_CONTAINER( table ), (GtkCallback) gtk_widget_destroy, NULL );
 
     if ( n_configs == 0 )
 	return;
@@ -556,87 +565,67 @@ void on_preference(GtkAction* act, gpointer data)
 static int on_track_info_received(xmmsv_t* value, void* user_data)
 {
     GtkBuilder* builder = (GtkBuilder*)user_data;
-    xmmsv_t *string_value, *int_value;
-    
+    TrackProperties tp;
     GtkWidget* w;
-    const char* keys[] = {
-        "title", "album", "artist", "comment", "mime", NULL
-    };
-    const char** key;
-    const char* val;
-    int ival;
-    /* xmmsc_result_propdict_foreach(res, dict_foreach, NULL); */
 
-    /* file name */
-
-    value = xmmsv_propdict_to_dict (value, NULL);
-    if ( xmmsv_dict_get( value, "url", &string_value ) )
+    get_track_properties( value, &tp );
+    if (tp.url && (!g_str_equal(tp.url, "" )))
     {
-	g_assert( xmmsv_get_string( string_value, &val ) );
         w = (GtkWidget*)gtk_builder_get_object(builder, "url");
-        if( g_str_has_prefix(val, "file://") )
+        if( g_str_has_prefix(tp.url, "file://") )
         {
             char* disp;
-	    gchar *decoded_val;
-
-	    decoded_val = xmmsv_url_to_string ( string_value );
-  
-	    val += 7; 
+	    gchar *decoded_val = g_uri_unescape_string( tp.url, NULL );
 	    /* skip file:// */
-	    if ( decoded_val != NULL )  
-	    {
-		disp = g_filename_display_name(decoded_val + 7 ) ;
-		g_free( decoded_val );
-	    }
-	    /* fallback, if decoding failed */
-	    else 
-		disp = g_filename_display_name(val + 7 ) ;		
+	    disp = g_filename_display_name( decoded_val + 7 ) ;
+	    g_free( decoded_val );
             gtk_entry_set_text(GTK_ENTRY(w), disp);
             g_free(disp);
         }
         else
-            gtk_entry_set_text(GTK_ENTRY(w), val);
+            gtk_entry_set_text(GTK_ENTRY(w), tp.url);
     }
-
-    for(key = keys; *key; ++key)
+    if (tp.title || tp.album || tp.artist || tp.comment || tp.mime) 
     {
-	if ( xmmsv_dict_get( value, *key, &string_value ) ) 
+	const char **key;
+	const char *val;
+	const char *key_values[] = {
+	    "title", tp.title, 
+	    "album", tp.album,
+	    "artist", tp.artist,
+	    "comment", tp.comment,
+	    "mime", tp.mime, 
+	    NULL
+	};
+
+	for ( key=key_values; *key; key += 2 ) 
 	{
-	    xmmsv_get_string( string_value, &val );
-            w = (GtkWidget*)gtk_builder_get_object(builder, *key);
-            if( GTK_IS_ENTRY(w) )
-                gtk_entry_set_text((GtkEntry*)w, val);
-            else if( GTK_IS_LABEL(w) )
-                gtk_label_set_text((GtkLabel*)w, val);
-        }
+	    /* check if string_property is non-empty */
+	    if (!(val = *(key + 1)))
+		continue;
+	    w = (GtkWidget*)gtk_builder_get_object(builder, *key);
+	    if( GTK_IS_ENTRY(w) )
+		gtk_entry_set_text((GtkEntry*)w, val);
+	    else if( GTK_IS_LABEL(w) )
+		gtk_label_set_text((GtkLabel*)w, val);
+	}
     }
-
+    
     /* size & bitrate */
-    if ( xmmsv_dict_get( value, "size", &int_value ) )
-    {
-        char buf[100];
+    if (tp.size){
+	char buf[100];
 	char *size_str;
-
-	xmmsv_get_int( int_value, &ival );
-	size_str = g_format_size_for_display( ival );
+	size_str = g_format_size_for_display( tp.size );
 	strcpy( buf, size_str );
 	g_free( size_str );
 
-	if ( xmmsv_dict_get( value, "bitrate", &int_value ) ) 
+	if (tp.bitrate) 
 	{
-            int len;
-	    int vbr = 0;
-	    xmmsv_get_int( int_value, &ival );
-	    
-	    len = strlen(buf);
-
-	    if ( xmmsv_dict_get( value, "isvbr", &int_value ) ) 
-		xmmsv_get_int( int_value, &vbr );
-
-	    g_snprintf(buf + len, 100 - len, " (%s%d Kbps%s)", _("Bitrate: "), ival/1000, vbr ? ", vbr" : "" );
+	    int len = strlen(buf);
+	    g_snprintf(buf + len, 100 - len, " (%s%d Kbps%s)", _("Bitrate: "), tp.bitrate/1000, tp.isvbr ? ", vbr" : "" );
 	}
-        w = (GtkWidget*)gtk_builder_get_object(builder, "size");
-        gtk_label_set_text((GtkLabel*)w, buf);
+	w = (GtkWidget*)gtk_builder_get_object(builder, "size");
+	gtk_label_set_text((GtkLabel*)w, buf);
     }
     xmmsv_unref( value );
     return TRUE;
@@ -711,11 +700,11 @@ void on_playlist_view_drag_data_received(GtkWidget          *widget,
     char** uris;
     GtkTreePath            *dest_path;
     GtkTreeViewDropPosition pos;
-    int insert_pos = gtk_tree_model_iter_n_children( gtk_tree_view_get_model ( playlist_view ), NULL );
+    int insert_pos = gtk_tree_model_iter_n_children( gtk_tree_view_get_model ( GTK_TREE_VIEW(playlist_view ) ), NULL );
 
     g_signal_stop_emission_by_name(playlist_view, "drag_data_received");
     
-    if ( gtk_tree_view_get_dest_row_at_pos( playlist_view, x, y, &dest_path, &pos ) ) 
+    if ( gtk_tree_view_get_dest_row_at_pos( GTK_TREE_VIEW(playlist_view), x, y, &dest_path, &pos ) ) 
     {
 	insert_pos = gtk_tree_path_get_indices( dest_path )[0] ;
 	if ( pos != GTK_TREE_VIEW_DROP_BEFORE )
@@ -984,11 +973,11 @@ void on_add_url( GtkMenuItem* item, gpointer user_data )
     }
     gtk_widget_destroy( dlg );
 }
+/* FIXME: This might be implemented in the future */
+/* static void on_add_from_mlib( GtkMenuItem* item, gpointer user_data ) */
+/* { */
 
-static void on_add_from_mlib( GtkMenuItem* item, gpointer user_data )
-{
-    /* FIXME: This might be implemented in the future */
-}
+/* } */
 
 void on_add_btn_clicked(GtkButton* btn, gpointer user_data)
 {
@@ -1159,106 +1148,130 @@ static void render_num( GtkTreeViewColumn* col, GtkCellRenderer* render,
 
 static int update_track( xmmsv_t *value, UpdateTrack* ut )
 {
-    const char *artist = NULL;
-    const char *album = NULL;
-    const char *title = NULL;
+    TrackProperties track_properties;
     gboolean current_track_updated;
-    xmmsv_t *string_value, *time_len_val;
-    int32_t time_len = 0;
     char time_buf[32];
-    /* g_debug("do update track: %d", ut->id); */
-
-    /* OK, now it's time to send the next request.
-     * This is inefficient, but it's used to overcome
-     * some design flaws of xmms2d.
-     * Some optimization can be done here by send about
-     * 10 requets or so at the same time. */
-    if( ! g_queue_is_empty(pending_update_tracks) )
-    {
-        xmmsc_result_t* res2;
-        UpdateTrack* ut = (UpdateTrack*)g_queue_pop_head(pending_update_tracks);
-
-        res2 = xmmsc_medialib_get_info( con, ut->id );
-        xmmsc_result_notifier_set_full( res2, (xmmsc_result_notifier_t)update_track, ut, (xmmsc_user_data_free_func_t)free_update_track );
-        xmmsc_result_unref( res2 );
-    }
-
+    gchar *guessed_title = NULL;
     if( xmmsv_is_error ( value ) ) {
         return FALSE;
     }
+
+    /* check if this update is valid: Maybe it was meanwhile canceled
+     * (for example by switching a long loading playlist */
+    if (!g_hash_table_lookup( update_tracks, ut )) 
+	return TRUE;
     
-
-    value = xmmsv_propdict_to_dict (value, NULL);
-
-    if ( xmmsv_dict_get( value, "artist", &string_value ) )
-	xmmsv_get_string( string_value, &artist );
-    if ( xmmsv_dict_get( value, "album", &string_value ) )
-	xmmsv_get_string( string_value, &album );    
-    if ( xmmsv_dict_get( value, "title", &string_value ) )
-	xmmsv_get_string( string_value, &title );    
-    if ( xmmsv_dict_get( value, "duration", &time_len_val ) )
-	xmmsv_get_int( time_len_val, &time_len);
-
-    timeval_to_str( time_len/1000, time_buf, G_N_ELEMENTS(time_buf) );
-
-
-    if( !title || g_str_equal( title, "" ) )
-	title =  xmmsv_media_dict_guess_title ( value );
+    /* valid update */
+    g_hash_table_remove( update_tracks, ut );
+    
+    if (!get_track_properties( value, &track_properties)) 
+	track_properties.title = guessed_title = guess_title_from_url( track_properties.url );
+    timeval_to_str( track_properties.duration/1000, time_buf, G_N_ELEMENTS(time_buf) );
 
     gtk_list_store_set( list_store, &ut->it,
-                        COL_ARTIST, artist,
-                        COL_ALBUM, album,
-                        COL_TITLE, title,
+                        COL_ARTIST, track_properties.artist,
+                        COL_ALBUM, track_properties.album,
+                        COL_TITLE, track_properties.title,
                         COL_LEN, time_buf, -1 );
 
-    current_track_updated = ut->it.user_data == cur_track_iter.user_data;
+    current_track_updated = ut->id == cur_track_id;
     if ( current_track_updated ) 
     {
 	/* send desktop notification if current track was updated */  
-	send_notifcation( artist, title );
+	send_notifcation( track_properties.artist, track_properties.title );
 	if( tray_icon ) 
 	{
-	    GString* tray_tooltip = create_window_title(artist, title, playback_status == XMMS_PLAYBACK_STATUS_PLAY);
+	    GString* tray_tooltip = create_window_title(track_properties.artist, track_properties.title, playback_status == XMMS_PLAYBACK_STATUS_PLAY);
 	    gtk_status_icon_set_tooltip( GTK_STATUS_ICON(tray_icon), tray_tooltip->str );
 	    g_string_free( tray_tooltip, TRUE );
 	}
 	
     }
+    g_slice_free(UpdateTrack, ut);
+    g_free( guessed_title );
+    return FALSE;
+}
 
-    xmmsv_unref( value );
+static gboolean get_track_properties (xmmsv_t *value, TrackProperties *properties)  
+{
+    /* traverse the dict of dict */
+    xmmsv_dict_iter_t *parent_it;
+
+    /* fallback for title: online streams often provide channel */
+    const char* channel = NULL;
+
+    /* default values: empty */
+    bzero( properties, sizeof(TrackProperties) );
+    
+    xmmsv_get_dict_iter (value, &parent_it);
+    while (xmmsv_dict_iter_valid (parent_it ) ) 
+    {
+	const char *key;
+	const char **val_str = NULL;
+	int32_t *val_int = NULL;
+	xmmsv_t *child_value;
+	xmmsv_dict_iter_t *child_it;
+	
+	/* get child dict */
+	xmmsv_dict_iter_pair (parent_it, &key, &child_value);
+
+	/* check type of property */
+	if (strcmp( key, "artist" ) == 0)
+	    val_str = &(properties->artist);
+	else if (strcmp( key, "album" ) == 0)
+	    val_str = &(properties->album);
+	else if (strcmp( key, "mime" ) == 0)
+	    val_str = &(properties->mime);
+	else if (strcmp( key, "comment" ) == 0)
+	    val_str = &(properties->comment);
+	else if (strcmp( key, "channel" ) == 0)
+	    val_str = &channel;	    	    
+	else if (strcmp( key, "url" ) == 0)
+	    val_str = &(properties->url);	    
+	else if (strcmp( key, "title" ) == 0) 
+	    val_str = &(properties->title);	    
+	else if (strcmp( key, "duration" ) == 0)
+	    val_int = &(properties->duration);
+	else if (strcmp( key, "isvbr" ) == 0)
+	    val_int = &(properties->isvbr);
+	else if (strcmp( key, "bitrate" ) == 0)
+	    val_int = &(properties->bitrate);
+	else if (strcmp( key, "size" ) == 0)
+	    val_int = &(properties->size);
+	
+	if (xmmsv_get_dict_iter (child_value, &child_it) && 
+	    xmmsv_dict_iter_valid (child_it) && (val_int || val_str) && 
+	    xmmsv_dict_iter_pair (child_it, NULL, &child_value)) {
+
+	    if (val_int != NULL) 
+		xmmsv_get_int( child_value, val_int);
+	    else 
+		xmmsv_get_string( child_value, val_str );    
+	}
+	xmmsv_dict_iter_next (parent_it);
+    }
+
+    if ((properties->title == NULL) || g_str_equal( properties->title, "" )) 
+    {
+	if (channel == NULL)
+	    return FALSE;
+	else
+	    properties->title = channel;
+    }
     return TRUE;
 }
 
 static void queue_update_track( uint32_t id, GtkTreeIter* it )
 {
     UpdateTrack* ut;
-    /* if it's already in the queue */
-    if( ! g_queue_is_empty( pending_update_tracks ) )
-    {
-        GList* l;
-        for( l = pending_update_tracks->head; l; l = l->next )
-        {
-            ut = (UpdateTrack*)l->data;
-            if( id == ut->id )
-                return;
-        }
-    }
-
+    xmmsc_result_t *res;
     ut = g_slice_new(UpdateTrack);
     ut->id = id;
     ut->it = *it;
-
-    if( g_queue_is_empty( pending_update_tracks ) )
-    {
-        xmmsc_result_t *res;
-        res = xmmsc_medialib_get_info( con, id );
-        xmmsc_result_notifier_set_full( res, (xmmsc_result_notifier_t)update_track, ut, (xmmsc_user_data_free_func_t)free_update_track );
-        xmmsc_result_unref( res );
-    }
-    else
-    {
-        g_queue_push_tail(pending_update_tracks, ut);
-    }
+    g_hash_table_insert( update_tracks, ut, ut );
+    res = xmmsc_medialib_get_info( con, id );
+    xmmsc_result_notifier_set_full( res, (xmmsc_result_notifier_t)update_track, ut, NULL );
+    xmmsc_result_unref( res );
 }
 
 static int on_playlist_content_received( xmmsv_t* value, GtkWidget* list_view )
@@ -1267,6 +1280,10 @@ static int on_playlist_content_received( xmmsv_t* value, GtkWidget* list_view )
     GtkTreeIter it;
     int pl_size = xmmsv_list_get_size( value );
     int i;
+
+    /* free prev. model filter */
+    if ((mf = gtk_tree_view_get_model(GTK_TREE_VIEW(playlist_view))))
+	g_object_unref(mf);
     
     list_store = gtk_list_store_new(N_COLS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT );
     mf = gtk_tree_model_filter_new(GTK_TREE_MODEL(list_store), NULL);
@@ -1274,34 +1291,26 @@ static int on_playlist_content_received( xmmsv_t* value, GtkWidget* list_view )
     gtk_tree_view_set_search_equal_func( GTK_TREE_VIEW(playlist_view), playlist_search_func, NULL, NULL );
     g_object_unref(list_store);
 
+    /* invalidate pending track updates */
     cancel_pending_update_tracks();
-
+    
     for ( i = 0; i < pl_size; i++ ) 
     {
         int32_t id;
 	xmmsv_t *current_value;
+	xmmsc_result_t *res;
         UpdateTrack* ut = g_slice_new(UpdateTrack);
 
-        gtk_list_store_append( list_store, &it );
 	xmmsv_list_get( value, i, &current_value );
 	xmmsv_get_int( current_value, &id );
-
-        gtk_list_store_set( list_store, &it, COL_ID, id, COL_WEIGHT, PANGO_WEIGHT_NORMAL, -1 );
-
+        gtk_list_store_insert_with_values ( list_store, &it, i, COL_ID, id, COL_WEIGHT, PANGO_WEIGHT_NORMAL, -1 );
+	
         ut->id = id;
         ut->it = it;
-        /* add this request to pending list */
-        g_queue_push_tail( pending_update_tracks, ut );
-    }
-
-    if( !g_queue_is_empty(pending_update_tracks) )
-    {
-        UpdateTrack* ut = (UpdateTrack*)g_queue_pop_head(pending_update_tracks);
-	xmmsc_result_t *res;
-        /* only send the first request, and subsequent requests will
-         * be sent after the previous one is returned. */
+	/* just insert dummy values so we can distinguish from NULL */
+	g_hash_table_insert( update_tracks, ut, ut );
         res = xmmsc_medialib_get_info( con, ut->id );
-        xmmsc_result_notifier_set_full( res, (xmmsc_result_notifier_t)update_track, ut, (xmmsc_user_data_free_func_t)free_update_track );
+        xmmsc_result_notifier_set_full( res, (xmmsc_result_notifier_t)update_track, ut, NULL );
         xmmsc_result_unref( res );
     }
 
@@ -1416,8 +1425,8 @@ static int on_playlist_loaded(xmmsv_t* value, gpointer user_data)
         g_free(cur_playlist);
         cur_playlist = g_strdup(name);
 
-        cur_track_iter.stamp = 0; /* invalidate the GtkTreeIter of currenyly played track. */
-        cur_track_id = -1;
+	/* invalidate currenyly played track id. */
+        cur_track_id = 0;
 
         /* update the menu */
         if( cur_playlist )
@@ -1613,14 +1622,22 @@ static int on_playlist_content_changed( xmmsv_t* value, void* user_data )
                 GtkTreeIter it;
 		xmmsv_get_int( int_value, &pos );
                 path = gtk_tree_path_new_from_indices( pos, -1 );
-                if( gtk_tree_model_get_iter( (GtkTreeModel*)list_store, &it, path ) )
+                if( gtk_tree_model_get_iter( (GtkTreeModel*)list_store, &it, path ) ) 
+		{
                     gtk_list_store_remove( list_store, &it );
+		    /* invalidate currently played track */
+		    xmmsv_dict_get( value, "id", &int_value );
+		    xmmsv_get_int( int_value, &id );
+		    if ( id == cur_track_id ) 
+			cur_track_id = 0;
+		}
                 gtk_tree_path_free( path );
 	    }
             break;
         case XMMS_PLAYLIST_CHANGED_CLEAR:
         {
             gtk_list_store_clear( list_store );
+	    cur_track_id = 0;
             break;
         }
         case XMMS_PLAYLIST_CHANGED_MOVE:
@@ -1713,44 +1730,28 @@ static int on_playback_playtime_changed( xmmsv_t* value, void* user_data )
 
 static int on_playback_track_loaded( xmmsv_t* value, void* user_data )
 {
-    const char* artist = NULL;
-    const char* title = NULL;
+    TrackProperties track_properties;
     const char* err;
-    
-    xmmsv_t *duration_value;
-    xmmsv_t *string_value;
+    const char *guessed_title;
     GString* window_title;
     
-    cur_track_duration = 0;    
-
     if (xmmsv_get_error (value, &err)) {
 	g_warning( "Server error: %s", err );
 	return TRUE;
     }
-    value = xmmsv_propdict_to_dict (value, NULL);
 
-    if( xmmsv_dict_get( value, "duration", &duration_value ) )
-	xmmsv_get_int( duration_value, &cur_track_duration );
+    if (!get_track_properties( value, &track_properties)) 
+	track_properties.title = guessed_title = guess_title_from_url( track_properties.url );
+    cur_track_duration = track_properties.duration;    
 
-    artist = NULL;
-    if ( xmmsv_dict_get( value, "artist", &string_value ) )
-	xmmsv_get_string( string_value, &artist );
-
-    if ( xmmsv_dict_get( value, "title", &string_value ) ) 
-	xmmsv_get_string( string_value, &title );
-
-    /* title guessing */
-    if ( !title || g_str_equal( title, "" ) )
-	title =  xmmsv_media_dict_guess_title ( value );
-
-    window_title = create_window_title(artist, title, playback_status == XMMS_PLAYBACK_STATUS_PLAY);
+    window_title = create_window_title(track_properties.artist, track_properties.title, playback_status == XMMS_PLAYBACK_STATUS_PLAY);
 
     gtk_window_set_title( GTK_WINDOW(main_win), window_title->str );
     
     if( tray_icon )
         gtk_status_icon_set_tooltip(GTK_STATUS_ICON(tray_icon), window_title->str);
 
-    send_notifcation( artist, title );
+    send_notifcation( track_properties.artist, track_properties.title );
 
     g_string_free( window_title, TRUE );
     xmmsv_unref( value );
@@ -1805,21 +1806,24 @@ static int on_playlist_pos_changed( xmmsv_t* val, void* user_data )
 	xmmsv_get_string( name_val, &name );
         if( name && cur_playlist && !strcmp(cur_playlist, name) )
         {
-	    xmmsv_t *pos_val;
-	    
-            if( xmmsv_dict_get( val, "position", &pos_val ) ) 
+	    xmmsv_t *pos_val, *int_value;
+
+            if( xmmsv_dict_get( val, "position", &pos_val ) && xmmsv_get_int( pos_val, &pos ) && (pos >= 0 ) ) 
             {
-		xmmsv_get_int( pos_val, &pos );
                 /* mark currently played song in the playlist with bold font. */
                 GtkTreePath* path = gtk_tree_path_new_from_indices( pos, -1 );
                 GtkTreeIter it;
                 if( gtk_tree_model_get_iter( GTK_TREE_MODEL(list_store), &it, path ) )
                 {
+		    GtkTreeIter cur_track_iter = get_current_track_iter();
+		    /* prev. current track */
                     if( cur_track_iter.stamp )
                         gtk_list_store_set(list_store, &cur_track_iter, COL_WEIGHT, PANGO_WEIGHT_NORMAL, -1);
-		    
                     gtk_list_store_set(list_store, &it, COL_WEIGHT, PANGO_WEIGHT_BOLD, -1);
-                    cur_track_iter = it;
+
+		    /* update id of currently played track */
+		    xmmsv_dict_get( val, "id", &int_value );
+		    xmmsv_get_int( int_value, &cur_track_id );
                 }
 		/* scroll to currently played song */
 		gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(playlist_view), path, NULL, FALSE, 0.0, 0.0 );
@@ -1830,9 +1834,9 @@ static int on_playlist_pos_changed( xmmsv_t* val, void* user_data )
     return TRUE;
 }
 
-
 void on_locate_cur_track(GtkAction* act, gpointer user_data)
 {
+    GtkTreeIter cur_track_iter = get_current_track_iter();
     if( cur_track_iter.stamp )
     {
         GtkTreeModelFilter* mf;
@@ -1854,6 +1858,24 @@ void on_locate_cur_track(GtkAction* act, gpointer user_data)
     }
 }
 
+static GtkTreeIter get_current_track_iter() 
+{
+    GtkTreeIter it, found_it = {0};
+    int count=0;
+    if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list_store), &it )) 
+    {
+	do {
+	    count++;
+	    gint weight;
+	    gtk_tree_model_get (GTK_TREE_MODEL(list_store), 
+				&it, COL_WEIGHT, &weight, -1);
+	    if (weight == PANGO_WEIGHT_BOLD) 
+		found_it = it;
+	} while (!found_it.stamp && gtk_tree_model_iter_next( GTK_TREE_MODEL(list_store), &it ));
+    }
+    return found_it;
+}
+
 static void get_channel_volumes( const char *key, xmmsv_t *value, void *user_data)
 {
     GSList** volumes = (GSList**)user_data;
@@ -1923,11 +1945,13 @@ static int on_playback_volume_changed( xmmsv_t* value, void* user_data )
     GSList* volumes = NULL, *l;
     uint32_t vol = 0;
 
-    /* FIXME: OSS4 and pulse audio disconnect when playback is stopped,
-     * and hence we will receive a empty result here when playback is stopped.
-     * We need to handle this more gracefully in the future. */
-
-    if( value != NULL )
+    if(xmmsv_is_error ( value )) 
+    {
+	const gchar *error_msg;
+	xmmsv_get_error( value, &error_msg );
+	g_warning( "%s: %s", __func__, error_msg );
+    }
+    else if (xmmsv_is_type( value, XMMSV_TYPE_DICT ))
     {
 	xmmsv_dict_foreach( value, get_channel_volumes, &volumes );
         for( l = volumes; l; l = l->next )
@@ -2049,9 +2073,10 @@ static void config_changed_foreach(const char *key, xmmsv_t *value, void* user_d
     }
 }
 
-static void on_configval_changed(xmmsv_t* value, void* user_data)
+static int on_configval_changed(xmmsv_t* value, void* user_data)
 {
     xmmsv_dict_foreach( value, config_changed_foreach, NULL );
+    return TRUE;
 }
 
 static void setup_xmms_callbacks()
@@ -2340,6 +2365,8 @@ int main (int argc, char *argv[])
     gtk_init(&argc, &argv);
 
     plugin_config_setup(con);
+
+    update_tracks =  g_hash_table_new( g_direct_hash, NULL );
        
     xmmsc_mainloop_gmain_init(con);
 
@@ -2359,8 +2386,6 @@ int main (int argc, char *argv[])
         gtk_main_iteration();
     gdk_display_sync(gtk_widget_get_display(main_win));
 
-    pending_update_tracks = g_queue_new();
-
     /* display currently active playlist */
     res = xmmsc_playlist_current_active(con);
     xmmsc_result_notifier_set_and_unref(res, on_playlist_get_active, NULL);
@@ -2374,7 +2399,8 @@ int main (int argc, char *argv[])
 
     gtk_main ();
 
-    g_queue_free(pending_update_tracks);
+    cancel_pending_update_tracks();
+    g_hash_table_unref ( update_tracks );
     save_config();
 
     return 0;
